// (c) Peter Kankowski, 2007. http://smallcode.weblogs.us
#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x0400
#include <limits.h>
#include <windows.h>
#include <windowsx.h>
#define NUM_OF(A) (sizeof(A)/sizeof((A)[0])) // Number of elements in array

INT g_pos = 0;
UINT g_page_height;
UINT g_item_height;

TCHAR* const lines[] = {
	TEXT("Jabberwocky"),
	TEXT(""),
	TEXT("'Twas brillig, and the slithy toves"),
	TEXT("Did gyre and gimble in the wabe:"),
	TEXT("All mimsy were the borogoves,"),
	TEXT("And the mome raths outgrabe."),
	TEXT(""),
	TEXT("Beware the Jabberwock, my son!"),
	TEXT("The jaws that bite, the claws that catch!"),
   	TEXT("Beware the Jubjub bird, and shun"),
   	TEXT("The frumious Bandersnatch!"),
	TEXT(""),
   	TEXT("He took his vorpal sword in hand:"),
   	TEXT("Long time the manxome foe he sought -"),
   	TEXT("So rested he by the Tumturn tree,"),
   	TEXT("And stood awhile in thought."),
	TEXT(""),
   	TEXT("And, as in uffish thought he stood;"),
   	TEXT("The Jabberwock, with eyes of flame,"),
   	TEXT("Came whiffling through the tulgey wood,"),
   	TEXT("And burbled as it came!"),
	TEXT(""),
   	TEXT("One, two! One, two! And through and through"),
   	TEXT("The vorpal blade went snicker-snackl"),
   	TEXT("He left it dead, and with its head"),
   	TEXT("He went galumphing back."),
	TEXT(""),
   	TEXT("And hast thou slain the Jabberwock?"),
   	TEXT("Come to my arms, my beamish boy!"),
   	TEXT("O frabjous day! Callooh! Callay!"),
   	TEXT("He chortled in his joy."),
	TEXT(""),
	TEXT("'Twas brillig, and the slithy toves"),
	TEXT("Did gyre and gimble in the wabe:"),
	TEXT("All mimsy were the borogoves,"),
	TEXT("And the mome raths outgrabe."),
	TEXT(""),
	TEXT("Lewis Carroll, 1855"),
};

// based on Raymon Chen's code, http://blogs.msdn.com/oldnewthing/archive/2003/08/07/54615.aspx
INT GetWheelLines(INT units) {
	/*
		* "Units" are the units of Delta. 120 units == 1 click
		* A "click" is one wheel movement. For each click, the control should be
		scrolled by lines_per_click lines (usually by 3 lines).
		If the units doesn't make a full click, they should be carried to the next click.
	*/
	//units /= 4; // Uncomment to test sub-detent mouse wheel handling
	const INT UNITS_PER_CLICK = WHEEL_DELTA;
	UINT lines_per_click;
	if(!SystemParametersInfo(SPI_GETWHEELSCROLLLINES, 0, &lines_per_click, 0))
		lines_per_click = 3;
	// If the control should be scrolled by more than 1 page, then scroll it by 1 page.
	// WHEEL_PAGESCROLL is handled here, too.
	if(lines_per_click > g_page_height)
		lines_per_click = g_page_height;
	static INT units_carryover = 0;
	// Get the number of lines to scroll by
	units += units_carryover;
	INT clicks = units / UNITS_PER_CLICK;
	// Save the carryover
	units_carryover = units % UNITS_PER_CLICK;
	return -clicks * (INT)lines_per_click;
}

void Scroll(HWND hWnd, INT new_pos) {
	INT max_pos = max(NUM_OF(lines) - g_page_height, 0);
	if(new_pos < 0)
		new_pos = 0;
	else if(new_pos > max_pos)
		new_pos = max_pos;

	// If the position was changed
	if(g_pos != new_pos) {
		// Full redraw. In a real program, you may wish to use ScrollWindowEx
		g_pos = new_pos;
		InvalidateRect(hWnd, NULL, TRUE);
	}
}

LRESULT CALLBACK WndProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam) {
	switch(msg) {
	case WM_SIZE: {
		HDC hdc = GetDC(hWnd);

		// Get item height
		TEXTMETRIC tm;
		GetTextMetrics(hdc, &tm);
		g_item_height = tm.tmHeight;

		// Calculate page height
		UINT window_height = HIWORD(lParam);
		g_page_height = window_height / g_item_height;

		ReleaseDC(hWnd, hdc);
		return 0;
		}

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;

	case WM_MOUSEWHEEL:
		Scroll(hWnd, g_pos + GetWheelLines(GET_WHEEL_DELTA_WPARAM(wParam)));
		return 0;

	case WM_PAINT: {
		PAINTSTRUCT ps;
		HDC hdc = BeginPaint(hWnd, &ps);
		SetBkMode(hdc, TRANSPARENT);

		// Draw lines
		INT y = 0;
		for(INT i = g_pos; i < NUM_OF(lines); ++i) {
			TextOut(hdc, 0, y, lines[i], lstrlen(lines[i]));
			y += g_item_height;
		}
		EndPaint(hWnd, &ps);
		return 0;
		}

	default:
		return DefWindowProc(hWnd, msg, wParam, lParam);
	}
}
	

INT WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
				   LPSTR lpCmdLine, INT nCmdShow) {
	UNREFERENCED_PARAMETER(hInstance);
	UNREFERENCED_PARAMETER(hPrevInstance);
	UNREFERENCED_PARAMETER(lpCmdLine);
	UNREFERENCED_PARAMETER(nCmdShow);
	HWND hWnd; MSG msg;
	static const TCHAR classname[] = TEXT("KankowskiWheelScroll");
	WNDCLASS wc = {0, (WNDPROC)WndProc, 0, 0,
		(HINSTANCE)NULL, (HICON)NULL, (HCURSOR)NULL, (HBRUSH)(COLOR_BTNFACE+1),
		NULL, classname};
	wc.hInstance = GetModuleHandle(NULL);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	if(!RegisterClass(&wc))
		return 1;
	hWnd = CreateWindow(classname, TEXT("Scroll with mouse wheel"), WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT, 400, 300,
		(HWND)NULL, (HMENU)NULL, (HINSTANCE)wc.hInstance, (LPVOID)NULL);
	if(!hWnd) return 1;
	ShowWindow(hWnd, SW_SHOW);
	while(GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return 0;
}
